# Contribution guidelines

It's great that people want to contribute to Tileson!
Here are some guidelines to make sure we keep this a stable library.

### 1. Every pull request must have an existing issue that is accepted
In short this means that I must be aware of an issue and accept it before it can be worked on. 
Reason being mainly to make sure the issue actually is a problem and that I retain full control of project and its direction. 
If the issue is about a potential new feature, it's important that you describe in detail why this feature is needed.

Also: If you are interested in fixing or developing on an existing issue you did not
create yourself, please notify me in the issue before you start working on it, so I 
can assign you the task. This is mainly to prevent that two people work on the 
same feature.

### 2. Every issue must have a unit test proving that it works
If you have either fixed a bug or created a new feature, you must prove that everything works
as intended by creating a test to prove that the logic works. Additionally, 
you must make sure that all the other tests still works after your change.

### 3. Issues that are not "fixing" anything will not be accepted by pull requests
A good example for this is naming violations that are inconsistent with the rest of 
the variable/function names. You are free to report them as anything else, but I
must eventually be the one fixing it. Reason being that any change is a potential 
breaking change, and I want to introduce as little breaking changes as possible. 
This especially for these cases where there is no technical problem fixed, but it
will result in a breaking change as a side effect.

### 4. Try to make your logic documentation friendly
Tileson uses Doxygen to generate documentation. You should leave a little 
description on the top of your function to describe how it works. Like this:
```c++
/*!
 * Adds to numbers together and returns the result
 * 
 * @param a The first number
 * @param b The second number
 * @return The result of a plus b.
 */
int addNumbers(int a, int b)
{
    return a + b;
}
```
The most important here is to have a description in the top explaining what the function does.
If the parameters speaks for themselves, you are free to leave these blank, if you want.

### 5. Make sure you do your changes to the code inside the `include` folder
`tileson.hpp` and `tileson_min.hpp` are generated by the code inside the `include` folder.
This means that any changes in the `.hpp` files will be overwritten, eventually, 
so make sure you do your changes in the right place.

### 6. Make sure the examples still works before you do a pull request.
The examples use the generated `tileson.hpp` file, so you need to make sure this file 
is up to date and compiles when adding new features. This can be done by simply going
to the `tools` folder and run the `amalgamate_script.bat` if you're on Windows, or
`amalgamate_script.sh` if you're on Linux. There will be added a MAC/OSX version of 
this script as well. If you are a MAC user and this script is not available, 
you can skip this step and leave it to me.

To build examples in a local environment, you'll have to set the CMake flags for 
`BUILD_EXAMPLES` and `CLEAR_TSON_CMAKE_CACHE` to `ON`. You might also need to reload 
the `CMakeLists.txt` file.

### 7. Update the Doxygen documentation
This is not something I strictly require for accepting a pull request, but it's nice
to keep in mind. You will need to have Doxygen installed on your local environment
to make this work, but you can do this by simply going into the `docs` folder with 
your console application of choice and run the `doxygen` command.